@@ -4,13 +4,31 @@
  * See Copyright for the status of this software.
  */
 
+#define XML_DEPRECATED
+
 #include <libxml/parser.h>
+#include <libxml/uri.h>
 #include <libxml/xmlreader.h>
 #include <libxml/xmlwriter.h>
 #include <libxml/HTMLparser.h>
 
 #include <string.h>
 
+static int
+testNewDocNode(void) {
+    xmlNodePtr node;
+    int err = 0;
+
+    node = xmlNewDocNode(NULL, NULL, BAD_CAST "c", BAD_CAST "");
+    if (node->children != NULL) {
+        fprintf(stderr, "empty node has children\n");
+        err = 1;
+    }
+    xmlFreeNode(node);
+
+    return err;
+}
+
 static int
 testStandaloneWithEncoding(void) {
     xmlDocPtr doc;
@@ -48,7 +66,8 @@ testUnsupportedEncoding(void) {
     xmlFreeDoc(doc);
 
     error = xmlGetLastError();
-    if (error->code != XML_ERR_UNSUPPORTED_ENCODING ||
+    if (error == NULL ||
+        error->code != XML_ERR_UNSUPPORTED_ENCODING ||
         error->level != XML_ERR_WARNING ||
         strcmp(error->message, "Unsupported encoding: #unsupported\n") != 0)
     {
@@ -78,6 +97,85 @@ testNodeGetContent(void) {
     return err;
 }
 
+static int
+testCFileIO(void) {
+    xmlDocPtr doc;
+    int err = 0;
+
+    /* Deprecated FILE-based API */
+    xmlRegisterInputCallbacks(xmlFileMatch, xmlFileOpen, xmlFileRead,
+                              xmlFileClose);
+    doc = xmlReadFile("test/ent1", NULL, 0);
+
+    if (doc == NULL) {
+        err = 1;
+    } else {
+        xmlNodePtr root = xmlDocGetRootElement(doc);
+
+        if (root == NULL || !xmlStrEqual(root->name, BAD_CAST "EXAMPLE"))
+            err = 1;
+    }
+
+    xmlFreeDoc(doc);
+    xmlPopInputCallbacks();
+
+    if (err)
+        fprintf(stderr, "xmlReadFile failed with FILE input callbacks\n");
+
+    return err;
+}
+
+#ifdef LIBXML_VALID_ENABLED
+static void
+testSwitchDtdExtSubset(void *vctxt, const xmlChar *name ATTRIBUTE_UNUSED,
+                       const xmlChar *externalId ATTRIBUTE_UNUSED,
+                       const xmlChar *systemId ATTRIBUTE_UNUSED) {
+    xmlParserCtxtPtr ctxt = vctxt;
+
+    ctxt->myDoc->extSubset = ctxt->_private;
+}
+
+static int
+testSwitchDtd(void) {
+    const char dtdContent[] =
+        "<!ENTITY test '<elem1/><elem2/>'>\n";
+    const char docContent[] =
+        "<!DOCTYPE doc SYSTEM 'entities.dtd'>\n"
+        "<doc>&test;</doc>\n";
+    xmlParserInputBufferPtr input;
+    xmlParserCtxtPtr ctxt;
+    xmlDtdPtr dtd;
+    xmlDocPtr doc;
+    xmlEntityPtr ent;
+    int err = 0;
+
+    input = xmlParserInputBufferCreateStatic(dtdContent,
+                                             sizeof(dtdContent) - 1,
+                                             XML_CHAR_ENCODING_NONE);
+    dtd = xmlIOParseDTD(NULL, input, XML_CHAR_ENCODING_NONE);
+
+    ctxt = xmlNewParserCtxt();
+    ctxt->_private = dtd;
+    ctxt->sax->externalSubset = testSwitchDtdExtSubset;
+    doc = xmlCtxtReadMemory(ctxt, docContent, sizeof(docContent) - 1, NULL,
+                            NULL, XML_PARSE_NOENT | XML_PARSE_DTDLOAD);
+    xmlFreeParserCtxt(ctxt);
+
+    ent = xmlGetDocEntity(doc, BAD_CAST "test");
+    if (ent->children->doc != NULL) {
+        fprintf(stderr, "Entity content should have NULL doc\n");
+        err = 1;
+    }
+
+    /* Free doc before DTD */
+    doc->extSubset = NULL;
+    xmlFreeDoc(doc);
+    xmlFreeDtd(dtd);
+
+    return err;
+}
+#endif /* LIBXML_VALID_ENABLED */
+
 #ifdef LIBXML_SAX1_ENABLED
 static int
 testBalancedChunk(void) {
@@ -386,13 +484,147 @@ testWriterClose(void){
 }
 #endif
 
+typedef struct {
+    const char *uri;
+    const char *base;
+    const char *result;
+} xmlRelativeUriTest;
+
+static int
+testBuildRelativeUri(void) {
+    xmlChar *res;
+    int err = 0;
+    int i;
+
+    static const xmlRelativeUriTest tests[] = {
+        {
+            "/a/b1/c1",
+            "/a/b2/c2",
+            "../b1/c1"
+        }, {
+            "a/b1/c1",
+            "a/b2/c2",
+            "../b1/c1"
+        }, {
+            "a/././b1/x/y/../z/../.././c1",
+            "./a/./b2/././b2",
+            "../b1/c1"
+        }, {
+            "file:///a/b1/c1",
+            "/a/b2/c2",
+            NULL
+        }, {
+            "/a/b1/c1",
+            "file:///a/b2/c2",
+            NULL
+        }, {
+            "a/b1/c1",
+            "/a/b2/c2",
+            NULL
+        }, {
+            "/a/b1/c1",
+            "a/b2/c2",
+            NULL
+        }, {
+            "http://example.org/a/b1/c1",
+            "http://example.org/a/b2/c2",
+            "../b1/c1"
+        }, {
+            "http://example.org/a/b1/c1",
+            "https://example.org/a/b2/c2",
+            NULL
+        }, {
+            "http://example.org/a/b1/c1",
+            "http://localhost/a/b2/c2",
+            NULL
+        }, {
+            "with space/x x/y y",
+            "with space/b2/c2",
+            "../x%20x/y%20y"
+        }, {
+            "with space/x x/y y",
+            "/b2/c2",
+            "with%20space/x%20x/y%20y"
+        }
+#if defined(_WIN32) || defined(__CYGWIN__)
+        , {
+            "\\a\\b1\\c1",
+            "\\a\\b2\\c2",
+            "../b1/c1"
+        }, {
+            "\\a\\b1\\c1",
+            "/a/b2/c2",
+            "../b1/c1"
+        }, {
+            "a\\b1\\c1",
+            "a/b2/c2",
+            "../b1/c1"
+        }, {
+            "file://server/a/b1/c1",
+            "\\\\?\\UNC\\server\\a\\b2\\c2",
+            "../b1/c1"
+        }, {
+            "file://server/a/b1/c1",
+            "\\\\server\\a\\b2\\c2",
+            "../b1/c1"
+        }, {
+            "file:///x:/a/b1/c1",
+            "x:\\a\\b2\\c2",
+            "../b1/c1"
+        }, {
+            "file:///x:/a/b1/c1",
+            "\\\\?\\x:\\a\\b2\\c2",
+            "../b1/c1"
+        }, {
+            "file:///x:/a/b1/c1",
+            "file:///y:/a/b2/c2",
+            NULL
+        }, {
+            "x:/a/b1/c1",
+            "y:/a/b2/c2",
+            "file:///x:/a/b1/c1"
+        }, {
+            "/a/b1/c1",
+            "y:/a/b2/c2",
+            NULL
+        }, {
+            "\\\\server\\a\\b1\\c1",
+            "a/b2/c2",
+            "file://server/a/b1/c1"
+        }
+#endif
+    };
+
+    for (i = 0; (size_t) i < sizeof(tests) / sizeof(tests[0]); i++) {
+        const xmlRelativeUriTest *test = tests + i;
+        const char *expect;
+
+        res = xmlBuildRelativeURI(BAD_CAST test->uri, BAD_CAST test->base);
+        expect = test->result ? test->result : test->uri;
+        if (!xmlStrEqual(res, BAD_CAST expect)) {
+            fprintf(stderr, "xmlBuildRelativeURI failed uri=%s base=%s "
+                    "result=%s expected=%s\n", test->uri, test->base,
+                    res, expect);
+            err = 1;
+        }
+        xmlFree(res);
+    }
+
+    return err;
+}
+
 int
 main(void) {
     int err = 0;
 
+    err |= testNewDocNode();
     err |= testStandaloneWithEncoding();
     err |= testUnsupportedEncoding();
     err |= testNodeGetContent();
+    err |= testCFileIO();
+#ifdef LIBXML_VALID_ENABLED
+    err |= testSwitchDtd();
+#endif
 #ifdef LIBXML_SAX1_ENABLED
     err |= testBalancedChunk();
 #endif
@@ -413,6 +645,7 @@ main(void) {
 #ifdef LIBXML_WRITER_ENABLED
     err |= testWriterClose();
 #endif
+    err |= testBuildRelativeUri();
 
     return err;
 }
